"""Tests for the `email` module."""

import pytest

from bolig_ping.data_models import Flat
from bolig_ping.email import compose_email


@pytest.mark.parametrize(
    argnames=["flats", "expected"],
    argvalues=[
        (
            [
                Flat(
                    url="https://some.url",
                    address="Some address",
                    price=1000,
                    num_rooms=3,
                    size=100,
                    monthly_fee=100,
                    year=2000,
                )
            ],
            (
                "[BoligPing] Found a new flat!",
                "Hi,\n\nI found a new flat that you might be interested in:\n\n"
                "<a href='https://some.url'>Some address</a>\n"
                "Price: 1,000 kr.\n"
                "Number of rooms: 3\n"
                "Size: 100 m²\n"
                "Monthly fee: 100 kr./md\n"
                "Year built: 2000\n\n"
                "Have a splendid day!\n\nBest regards,\nBoligPing",
            ),
        ),
        (
            [
                Flat(
                    url="https://some.url",
                    address="Some address",
                    price=1000,
                    num_rooms=3,
                    size=100,
                    monthly_fee=100,
                    year=2000,
                ),
                Flat(
                    url="https://another.url",
                    address="Another address",
                    price=2000,
                    num_rooms=4,
                    size=200,
                    monthly_fee=200,
                    year=2001,
                ),
            ],
            (
                "[BoligPing] Found 2 new flats!",
                "Hi,\n\nI found some new flats that you might be interested in:\n\n"
                "<a href='https://some.url'>Some address</a>\n"
                "Price: 1,000 kr.\n"
                "Number of rooms: 3\n"
                "Size: 100 m²\n"
                "Monthly fee: 100 kr./md\n"
                "Year built: 2000\n\n"
                "<a href='https://another.url'>Another address</a>\n"
                "Price: 2,000 kr.\n"
                "Number of rooms: 4\n"
                "Size: 200 m²\n"
                "Monthly fee: 200 kr./md\n"
                "Year built: 2001\n\n"
                "Have a splendid day!\n\nBest regards,\nBoligPing",
            ),
        ),
    ],
    ids=["one_flat", "two_flats"],
)
def test_compose_email(flats: list[Flat], expected: str) -> None:
    """Test that an email is composed."""
    email = compose_email(flats=flats)
    assert email == expected



"""Data models used in the project."""

from pydantic import BaseModel


class SearchQuery(BaseModel):
    """A flat search query."""

    cities: list[str]
    min_price: int
    max_price: int
    min_rooms: int
    max_rooms: int
    min_size: int
    max_size: int
    queries: list[str]

    def get_url(self) -> str:
        """Get the URL for the search query.

        Returns:
            The URL for the search query.
        """
        url = "https://www.boligsiden.dk/by/{}/tilsalg/ejerlejlighed".format(
            ",".join(self.cities)
        )
        arguments: dict[str, str | int] = dict(
            priceMin=self.min_price,
            priceMax=self.max_price,
            numberOfRoomsMin=self.min_rooms,
            numberOfRoomsMax=self.max_rooms,
            areaMin=self.min_size,
            areaMax=self.max_size,
        )
        if self.queries:
            arguments["text"] = ",".join(self.queries)
        url += "?" + "&".join(f"{key}={value}" for key, value in arguments.items())
        return url


class Flat(BaseModel):
    """A flat listing."""

    url: str
    address: str
    price: int | None
    num_rooms: int | None
    size: int | None
    monthly_fee: int | None
    year: int | None

    def __hash__(self) -> int:
        """Get the hash of the flat.

        Returns:
            The hash of the flat.
        """
        return hash(self.url)

    def to_html(self) -> str:
        """Get the flat as an HTML string.

        Returns:
            The flat as an HTML string.
        """
        html = "\n".join(
            [
                f"<a href='{self.url}'>{self.address}</a>",
                f"Price: {self.price:,} kr.",
                f"Number of rooms: {self.num_rooms}",
                f"Size: {self.size} m²",
                f"Monthly fee: {self.monthly_fee:,} kr./md",
                f"Year built: {self.year}",
            ]
        )
        return html



"""Scraping flats available satisfying the given criteria."""

import logging
import re

from selenium.common.exceptions import NoSuchElementException, TimeoutException
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webelement import WebElement
from tqdm.auto import tqdm

from .data_models import Flat, SearchQuery
from .webdriver import Webdriver

logger = logging.getLogger(__package__)


def scrape_results(search_query: SearchQuery) -> list[Flat]:
    """Scrape the results of a flat search query.

    Args:
        search_query:
            The search query to scrape results for.

    Returns:
        A list of flats that satisfy the search query.

    Raises:
        HTTPError:
            If there was an error in the HTTP request.
    """
    url = search_query.get_url()
    logger.info(f"Fetching URL {url!r}...")
    webdriver = Webdriver().load(url=url)

    # Close the cookie banner
    logger.info("Closing cookie banner...")
    webdriver.click_element_or_ignore(
        xpath="//button[@id='didomi-notice-disagree-button']"
    )

    # Get the number of pages
    logger.info("Getting number of results...")
    try:
        num_results_elt = webdriver.find_element(
            "//h1[contains(concat(' ', normalize-space(@class), ' '), ' text-xl ')]"
        )
    except NoSuchElementException:
        raise ValueError(
            f"Could not find number of results for search query: {search_query}"
        )
    except TimeoutException:
        raise TimeoutError(
            "Timed out while trying to fetch number of results for search query: "
            f"{search_query}."
        )

    num_results_match = re.search(r"[0-9]+", num_results_elt.text)
    if num_results_match is None:
        raise ValueError("Could not find number of results.")
    num_results = int(num_results_match.group())

    # Extract the flats from the first page
    logger.info("Scraping first page...")
    results = webdriver.find_elements(
        "//div[@data-testid='case-list-card' and "
        "contains(concat(' ', normalize-space(@class), ' '), ' shadow-card ')]"
    )
    flats = [get_flat_from_result(result=result) for result in results]

    # Scrape the remaining pages
    with tqdm(desc="Scraping pages") as pbar:
        # Update the progress bar
        pbar.total = num_results
        pbar.update(len(flats))

        # Calculate the number of pages
        num_pages = num_results // 50
        if num_results % 50 != 0:
            num_pages += 1

        # Iterate over the remaining pages
        for _ in range(num_pages - 1):
            # Go to next page page
            webdriver.click_element_or_ignore(
                xpath="//ul[@role='navigation']//a[@role='button' and @rel='next']"
            )

            # Get the results, where we keep trying in case the page hasn't changed
            # correctly
            num_flats = len(flats)
            new_flats = []
            num_attempts = 3
            while len(flats) == num_flats:
                # Get the results
                results = webdriver.find_elements(
                    "//div["
                    "@data-testid='case-list-card' and contains("
                    "concat(' ', normalize-space(@class), ' '), ' shadow-card ')"
                    "]"
                )
                new_flats = [get_flat_from_result(result=result) for result in results]
                flats.extend(new_flats)
                flats = list(set(flats))

                # Monitor the number of attempts and raise an error if we can't change
                # the page
                num_attempts -= 1
                if num_attempts == 0:
                    raise ValueError("Could not change page.")
            pbar.update(len(new_flats))

        # Ensure that the progress bar is at 100% at the end
        pbar.n = pbar.total

    return flats


def get_flat_from_result(result: WebElement) -> Flat:
    """Get a flat from a result.

    Args:
        result:
            The result to get the flat from.

    Returns:
        The flat from the result.

    Raises:
        ValueError:
            If the result could not be parsed.
    """
    # Extract URL
    candidate_urls = [
        url.get_attribute("href") or ""
        for url in result.find_elements(By.XPATH, ".//a")
        if "viderestilling" in (url.get_attribute("href") or "")
    ]
    if len(candidate_urls) == 0:
        raise ValueError("Could not find URL in result.")
    url = candidate_urls[0].split("?")[0]
    if not url.startswith("https"):
        url = "https://boligsiden.dk" + url

    # Extract address
    try:
        address = result.find_element(
            By.XPATH,
            ".//div["
            "contains(concat(' ', normalize-space(@class), ' '), ' bg-black ')"
            "]//div["
            "contains(concat(' ', normalize-space(@class), ' '), ' font-black ')"
            "]",
        ).text.replace("\n", " ")
    except NoSuchElementException:
        raise ValueError(f"Could not find address in result: {result.text}")

    # Extract span values
    all_span_values = {span.text for span in result.find_elements(By.XPATH, ".//span")}
    span_regexes = dict(
        price=r"kr\.$",
        size=r"[0-9]+ m²",
        num_rooms=r"[0-9]+ Vær",
        monthly_fee=r"Ejerudg.*kr\.?/md",
        year=r"Opført.*[0-9]{4}",
    )
    values = dict()
    for span_value in all_span_values:
        for name, regex in span_regexes.items():
            if re.search(pattern=regex, string=span_value) is not None:
                values[name] = extract_number(span_value)

    return Flat(url=url, address=address, **values)


def extract_number(text: str) -> int | None:
    """Extract the number from a string.

    Args:
        text:
            The string containing the number.

    Returns:
        The number, or None if no number was found.
    """
    match = re.search(r"[0-9][0-9\.]*", text)
    if match is None:
        return None
    match_str = match.group().replace(".", "")
    if match_str == "":
        return None
    return int(match_str)



"""Tests for the `cache` module."""

from collections.abc import Generator
from pathlib import Path

import pytest

from bolig_ping.cache import store_to_cache
from bolig_ping.data_models import Flat


class TestStoreToCache:
    """Tests for the store_to_cache function."""

    @pytest.fixture(scope="class")
    def flat(self) -> Generator[Flat, None, None]:
        """Return a Flat object."""
        yield Flat(
            url="https://some.url",
            address="Some address",
            price=1000,
            num_rooms=3,
            size=100,
            monthly_fee=100,
            year=2000,
        )

    @pytest.fixture(scope="class")
    def another_flat(self) -> Generator[Flat, None, None]:
        """Return another Flat object."""
        yield Flat(
            url="https://another.url",
            address="Another address",
            price=2000,
            num_rooms=4,
            size=200,
            monthly_fee=200,
            year=2001,
        )

    def test_flat_is_stored(self, flat: Flat) -> None:
        """Test that a flat is stored."""
        cache_path = Path(".test_cache")
        store_to_cache(flats=[flat], email="no-email", cache_path=cache_path)
        with cache_path.open() as file:
            assert file.read() == '{"id": "some.url", "email": "no-email"}\n'
        cache_path.unlink()

    def test_multiple_flats_are_stored(self, flat: Flat, another_flat: Flat) -> None:
        """Test that multiple flats are stored."""
        cache_path = Path(".test_cache")
        store_to_cache(
            flats=[flat, another_flat], email="no-email", cache_path=cache_path
        )
        with cache_path.open() as file:
            assert (
                file.read()
                == '{"id": "some.url", "email": "no-email"}\n'
                + '{"id": "another.url", "email": "no-email"}\n'
            )
        cache_path.unlink()

    def test_no_duplicates_are_stored(self, flat: Flat) -> None:
        """Test that no duplicates are stored."""
        cache_path = Path(".test_cache")
        store_to_cache(flats=[flat, flat], email="no-email", cache_path=cache_path)
        with cache_path.open() as file:
            assert file.read() == '{"id": "some.url", "email": "no-email"}\n'
        store_to_cache(flats=[flat], email="no-email", cache_path=cache_path)
        with cache_path.open() as file:
            assert file.read() == '{"id": "some.url", "email": "no-email"}\n'
        cache_path.unlink()



"""Command line interface for the project."""

import logging
import os

import click

from .cache import remove_cached_flats, store_to_cache
from .data_models import SearchQuery
from .email import compose_email, send_email
from .scraper import scrape_results

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s ⋅ %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
)
logger = logging.getLogger(__package__)


@click.command("bolig_ping")
@click.option(
    "--city",
    "-c",
    type=str,
    multiple=True,
    required=True,
    help="The city to search for apartments in.",
)
@click.option(
    "--min-price",
    type=int,
    default=0,
    show_default=True,
    help="The minimum price of the apartment, in DKK.",
)
@click.option(
    "--max-price",
    type=int,
    default=int(1e9),
    show_default=True,
    help="The maximum price of the apartment, in DKK.",
)
@click.option(
    "--min-rooms",
    type=int,
    default=1,
    show_default=True,
    help="The minimum number of rooms in the apartment.",
)
@click.option(
    "--max-rooms",
    type=int,
    default=int(1e9),
    show_default=True,
    help="The maximum number of rooms in the apartment.",
)
@click.option(
    "--min-size",
    type=int,
    default=0,
    show_default=True,
    help="The minimum size of the apartment, in square meters.",
)
@click.option(
    "--max-size",
    type=int,
    default=int(1e9),
    show_default=True,
    help="The maximum size of the apartment, in square meters.",
)
@click.option(
    "--email",
    type=str,
    default=None,
    show_default=True,
    help="Email address to send the notification to, or None to print to stdout.",
)
@click.option("--query", "-q", multiple=True, help="A query to filter the results by.")
def main(
    city: list[str],
    min_price: int,
    max_price: int,
    min_rooms: int,
    max_rooms: int,
    min_size: int,
    max_size: int,
    query: list[str],
    email: str | None,
) -> None:
    """Search for flats in Denmark."""
    cities = [
        c.replace(" ", "-")
        .replace("ø", "oe")
        .replace("æ", "ae")
        .replace("å", "aa")
        .lower()
        for c in city
    ]
    search_query = SearchQuery(
        cities=cities,
        min_price=min_price,
        max_price=max_price,
        min_rooms=min_rooms,
        max_rooms=max_rooms,
        min_size=min_size,
        max_size=max_size,
        queries=query,
    )
    flats = scrape_results(search_query=search_query)
    flats = remove_cached_flats(flats=flats, email=email or "no-email")
    logger.info(f"Found {len(flats)} new flats that satisfy the search query.")
    if flats:
        if email is not None:
            subject, contents = compose_email(flats=flats)
            send_email(
                from_email=os.environ["GMAIL_EMAIL"],
                password=os.environ["GMAIL_PASSWORD"],
                to_email=email,
                subject=subject,
                contents=contents,
            )
            logger.info(f"Sent the flats to {email}.")
        else:
            logger.info(
                "No email provided, so printing the flats here:\n\n"
                + "\n\n".join(flat.to_html() for flat in flats)
            )
        store_to_cache(flats=flats, email=email or "no-email")


if __name__ == "__main__":
    main()



"""Sending emails with flat listings."""

import yagmail
from dotenv import load_dotenv

from .data_models import Flat

load_dotenv()


def compose_email(flats: list[Flat]) -> tuple[str, str]:
    """Compose an email with the given flats.

    Args:
        flats:
            The flats to compose the email with.

    Returns:
        A pair (subject, contents) for the email.
    """
    match len(flats):
        case 0:
            raise ValueError("Cannot compose an email with no flats.")
        case 1:
            subject = "[BoligPing] Found a new flat!"
            contents = "Hi,\n\nI found a new flat that you might be interested in:\n\n"
        case _:
            subject = f"[BoligPing] Found {len(flats)} new flats!"
            contents = (
                "Hi,\n\nI found some new flats that you might be interested in:\n\n"
            )
    contents += "\n\n".join(flat.to_html() for flat in flats)
    contents += "\n\nHave a splendid day!\n\nBest regards,\nBoligPing"
    return subject, contents


def send_email(
    from_email: str, password: str, to_email: str, subject: str, contents: str
) -> None:
    """Send an email with the given contents.

    Args:
        from_email:
            The email to send the email from.
        password:
            The password for the from email.
        to_email:
            The email to send the email to.
        subject:
            The subject of the email.
        contents:
            The contents of the email.
    """
    yagmail.SMTP(user=from_email, password=password).send(
        to=to_email, subject=subject, contents=contents
    )



"""Checks related to the .env file in the repository.

Usage:
    uv run src/scripts/fix_dot_env_file.py [--non-interactive]
"""

from pathlib import Path

import click

# List of all the environment variables that are desired
DESIRED_ENVIRONMENT_VARIABLES = dict(
    GIT_NAME="Enter your full name, to be shown in Git commits:\n> ",
    GIT_EMAIL="Enter your email, as registered on your Github account:\n> ",
)


@click.command()
@click.option(
    "--non-interactive",
    is_flag=True,
    default=False,
    help="If set, the script will not ask for user input.",
)
def fix_dot_env_file(non_interactive: bool) -> None:
    """Ensures that the .env file exists and contains all desired variables.

    Args:
        non_interactive:
            If set, the script will not ask for user input.
    """
    env_path = Path(".env")
    name_and_email_path = Path(".name_and_email")

    # Ensure that the files exists
    env_path.touch(exist_ok=True)
    name_and_email_path.touch(exist_ok=True)

    # Extract all the lines in the files
    env_file_lines = [
        line for line in env_path.read_text().splitlines(keepends=False) if line.strip()
    ]
    name_and_email_file_lines = name_and_email_path.read_text().splitlines(
        keepends=False
    )

    # Extract all the environment variables in the files
    env_vars = {line.split("=")[0]: line.split("=")[1] for line in env_file_lines}
    name_and_email_vars = {
        line.split("=")[0]: line.split("=")[1] for line in name_and_email_file_lines
    }

    desired_env_vars = DESIRED_ENVIRONMENT_VARIABLES

    # For each of the desired environment variables, check if it exists in the .env
    # file
    env_vars_missing = [
        env_var for env_var in desired_env_vars.keys() if env_var not in env_vars
    ]

    # Create all the missing environment variables
    with env_path.open("a") as f:
        for env_var in env_vars_missing:
            value = ""

            if env_var in name_and_email_vars:
                value = name_and_email_vars[env_var]

            if value == "" and not non_interactive:
                value = input(desired_env_vars[env_var])

            f.write(f"{env_var}={value}\n")

    # Remove the name and email file
    name_and_email_path.unlink()


if __name__ == "__main__":
    fix_dot_env_file()



"""Tests for the `scraper` module."""

import pytest

from bolig_ping.scraper import extract_number


@pytest.mark.parametrize(
    argnames=["text", "expected"],
    argvalues=[
        ("1.000 kr.", 1000),
        ("1.000.000 kr.", 1000000),
        ("I alt 1.000.000 kr.", 1000000),
        ("I alt\n\n1.000.000 kr.", 1000000),
    ],
    ids=["Thousands", "Millions", "With text", "With text and newlines"],
)
def test_extract_number(text: str, expected: int) -> None:
    """Test the `extract_number` function."""
    assert extract_number(text=text) == expected



"""Tests for the `data_models` module."""

from collections.abc import Generator

import pytest

from bolig_ping.data_models import Flat, SearchQuery


class TestSearchQuery:
    """Tests for the `SearchQuery` data model."""

    @pytest.fixture(scope="class")
    def search_query(self) -> Generator[SearchQuery, None, None]:
        """Return a `SearchQuery` instance."""
        yield SearchQuery(
            cities=["københavn-n"],
            min_price=100,
            max_price=200,
            min_rooms=3,
            max_rooms=5,
            min_size=50,
            max_size=100,
            queries=["badekar"],
        )

    def test_create_search_query(self, search_query: SearchQuery) -> None:
        """Test creating a `SearchQuery` instance."""
        assert search_query.cities == ["københavn-n"]
        assert search_query.min_price == 100
        assert search_query.max_price == 200
        assert search_query.min_rooms == 3
        assert search_query.min_size == 50
        assert search_query.queries == ["badekar"]

    def test_get_url(self, search_query: SearchQuery) -> None:
        """Test the `get_url` method."""
        url = search_query.get_url()
        assert url == (
            "https://www.boligsiden.dk/by/københavn-n/tilsalg/ejerlejlighed"
            "?priceMin=100&priceMax=200&numberOfRoomsMin=3&numberOfRoomsMax=5"
            "&areaMin=50&areaMax=100&text=badekar"
        )


class TestFlat:
    """Tests for the `Flat` data model."""

    @pytest.fixture(scope="class")
    def flat(self) -> Generator[Flat, None, None]:
        """Return a `Flat` instance."""
        yield Flat(
            url="https://some.url",
            address="Some address",
            price=1000,
            num_rooms=3,
            size=100,
            monthly_fee=100,
            year=2000,
        )

    def test_create_flat(self, flat: Flat) -> None:
        """Test creating a `Flat` instance."""
        assert flat.url == "https://some.url"
        assert flat.address == "Some address"
        assert flat.price == 1000
        assert flat.num_rooms == 3
        assert flat.size == 100
        assert flat.monthly_fee == 100
        assert flat.year == 2000

    def test_hash(self, flat: Flat) -> None:
        """Test the `__hash__` method."""
        assert hash(flat) == hash(flat.url)



"""Tests for the `cli` module."""

import pytest
from click.testing import CliRunner

from bolig_ping.cli import main


@pytest.mark.parametrize(
    argnames=["cli_args"],
    argvalues=[
        ("--city københavn-n --max-price 100",),
        ("--city københavn-n --city brøndby --max-price 100",),
    ],
    ids=["One city", "Multiple cities"],
)
def test_main(cli_args: str) -> None:
    """Test the main function."""
    runner = CliRunner()
    result = runner.invoke(cli=main, args=cli_args)
    assert result.exit_code == 0



"""Webdriver used for scraping."""

import logging
from time import sleep

from selenium import webdriver
from selenium.common import WebDriverException
from selenium.common.exceptions import (
    NoSuchElementException,
    SessionNotCreatedException,
    TimeoutException,
)
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webelement import WebElement

logger = logging.getLogger(__package__)


class Webdriver:
    """A WebDriver to use for scraping."""

    def __init__(
        self,
        num_attempts: int = 3,
        timeout: int = 10,
        sleep_time: int = 3,
        headless: bool = True,
        **_,
    ) -> None:
        """Initialise the WebDriver.

        Args:
            num_attempts:
                The maximum number of attempts to get a page.
            timeout:
                The maximum time to wait for a page to load, in seconds.
            sleep_time:
                The time to sleep between attempts, in seconds.
            headless:
                Whether to run the WebDriver in headless mode.
        """
        self.num_attempts = num_attempts
        self.timeout = timeout
        self.sleep_time = sleep_time
        self.headless = headless
        self.driver = self.get_driver()

    @property
    def html(self) -> str:
        """Get the HTML of the current page.

        Returns:
            The HTML of the current page.
        """
        return self.driver.page_source

    def get_driver(self) -> webdriver.Chrome:
        """Get the WebDriver options.

        Returns:
            The WebDriver options.
        """
        options = webdriver.ChromeOptions()

        chrome_arguments = [
            "--no-sandbox",
            "--remote-debugging-port=9222",
            "--remote-debugging-pipe",
            "--autoplay-policy=no-user-gesture-required",
            "--no-first-run",
            "--disable-gpu",
            "--use-fake-ui-for-media-stream",
            "--use-fake-device-for-media-stream",
            "--disable-sync",
            "--disable-dev-shm-usage",
            "--start-maximized",
        ]
        for argument in chrome_arguments:
            options.add_argument(argument=argument)

        if self.headless:
            options.add_argument("--headless=new")

        attempts = 3
        for _ in range(attempts):
            try:
                driver = webdriver.Chrome(options=options)
                break
            except SessionNotCreatedException:
                logger.error("Could not create a new session. Trying again...")
                sleep(3)
        else:
            raise ConnectionError("Could not create a new session.")

        driver.set_page_load_timeout(time_to_wait=self.timeout)
        return driver

    def load(self, url: str) -> "Webdriver":
        """Load the DOM of a web page.

        Args:
            url:
                The URL of the web page.

        Raises:
            ConnectionError:
                If the page couldn't load.
        """
        for _ in range(self.num_attempts):
            try:
                self.driver.get(url=url)
                sleep(self.sleep_time)
                return self
            except (WebDriverException, TimeoutException) as e:
                logger.error(
                    f"{type(e).__name__} occurred while fetching the web page {url!r}. "
                    f"Retrying..."
                )
                sleep(self.sleep_time)
                return self
        else:
            raise ConnectionError(f"Could not load website {url}.")

    def find_element(self, xpath: str) -> WebElement:
        """Find an element by XPath.

        Args:
            xpath:
                The XPath of the element.

        Returns:
            The element found by the XPath, or None if not found.
        """
        return self.driver.find_element(by=By.XPATH, value=xpath)

    def find_elements(self, xpath: str) -> list[WebElement]:
        """Find elements by XPath.

        Args:
            xpath:
                The XPath of the elements.

        Returns:
            The elements found by the XPath.
        """
        return self.driver.find_elements(by=By.XPATH, value=xpath)

    def click_element_or_ignore(self, xpath: str) -> None:
        """Click an element by XPath, or ignore if not found.

        Args:
            xpath:
                The XPath of the element.
        """
        try:
            self.find_element(xpath=xpath).click()
            sleep(3)
        except NoSuchElementException:
            pass



"""Cache to store already sent flats."""

import json
from pathlib import Path

from .data_models import Flat


def store_to_cache(
    flats: list[Flat], email: str, cache_path: Path = Path(".boligping_cache")
) -> None:
    """Store the flats to the cache.

    Args:
        flats:
            The flats to store in the cache.
        email:
            The receiver of the flats.
        cache_path (optional):
            The path to the cache file. Defaults to ".boligping_cache".
    """
    flats = remove_cached_flats(flats=flats, email=email, cache_path=cache_path)
    added_flats: set[tuple[str, str]] = set()
    with cache_path.open("a") as file:
        for flat in flats:
            flat_id = flat.url.split("/")[-1]
            if (flat_id, email) in added_flats:
                continue
            flat_json = json.dumps(dict(id=flat_id, email=email))
            file.write(f"{flat_json}\n")
            added_flats.add((flat_id, email))


def remove_cached_flats(
    flats: list[Flat], email: str, cache_path: Path = Path(".boligping_cache")
) -> list[Flat]:
    """Remove the cached flats from the list of flats.

    Args:
        flats:
            The flats to remove the cached flats from.
        email:
            The receiver of the flats.
        cache_path (optional):
            The path to the cache file. Defaults to ".boligping_cache".

    Returns:
        The flats without the cached flats.
    """
    cache_path.touch(exist_ok=True)
    with cache_path.open() as file:
        for line in file:
            json_data = json.loads(line)
            if json_data["email"] == email:
                flats = [
                    flat for flat in flats if flat.url.split("/")[-1] != json_data["id"]
                ]
    return flats
